
typedef struct {
  double x,y;
  double NormalLen;
} TModelLinePoint;

typedef struct {
  double TotalLen;
  double fx,fy;
  u32 PointsCount;
  TModelLinePoint *pPoints;
} TModelLine;

typedef struct {
  UnicodeChar wc;
  u32 LinesCount;
  TModelLine *pLines;
} TModel;

static u32 ModelsCount;
static TModel *pModels=NULL;

static void Model_Draw(const TModel *pModel,CglCanvas *pcan,double ofsx,double ofsy,double ratio,u16 col);
static void Models_AddRefer(const TModel *pSrcModel,UnicodeChar wc);

#include "OSK_HW_Model_ModelLine.h"
#include "OSK_HW_Model_ModelResults.h"
#include "OSK_HW_Model_ModelsLoadSave.h"
#include "OSK_HW_Model_UserModel.h"

// ---------------------------------------------------------------------------

static void Model_Free(TModel *pmdl)
{
  pmdl->wc=0;
  for(u32 lidx=0;lidx<pmdl->LinesCount;lidx++){
    TModelLine *pline=&pmdl->pLines[lidx];
    pline->TotalLen=0;
    pline->fx=0;
    pline->fy=0;
    for(u32 pidx=0;pidx<pline->PointsCount;pidx++){
      TModelLinePoint *ppoi=&pline->pPoints[pidx];
      ppoi->x=0;
      ppoi->y=0;
    }
    pline->PointsCount=0;
    if(pline->pPoints!=NULL){
      safefree(&MM_SKKOSKHW,pline->pPoints); pline->pPoints=NULL;
    }
  }
  pmdl->LinesCount=0;
  if(pmdl->pLines!=NULL){
    safefree(&MM_SKKOSKHW,pmdl->pLines); pmdl->pLines=NULL;
  }
}

static bool Model_Fit(const TModel *pSrcModel,TModel *pDstModel)
{
  double minx=10000,miny=10000;
  double maxx=-10000,maxy=-10000;

#define chk(x,y) { \
  if(x<minx) minx=x; \
  if(maxx<x) maxx=x; \
  if(y<miny) miny=y; \
  if(maxy<y) maxy=y; \
}

  for(u32 lidx=0;lidx<pSrcModel->LinesCount;lidx++){
    TModelLine *pline=&pSrcModel->pLines[lidx];
    double px=pline->fx,py=pline->fy;
    for(u32 pidx=0;pidx<pline->PointsCount;pidx++){
      TModelLinePoint *ppoi=&pline->pPoints[pidx];
      px+=ppoi->x;
      py+=ppoi->y;
      chk(px,py);
    }
  }
  
  bool KogakiFlag=false;
  _consolePrintf("%f,%f.\n",miny,maxy);
  if((80<=miny)&&(120<=maxy)){
    KogakiFlag=true;
  }

  _consolePrintf("ModelFit: min=%f, %f, max=%f, %f.\n",minx,miny,maxx,maxy);
  double ratiox=1,ratioy=1;
  double ofsx,ofsy;
  {
    double w=maxx-minx;
    double h=maxy-miny;
    double ratioadd=0;
    if(w!=0) ratiox=256/w;
    if(h!=0) ratioy=256/h;
    if(ratiox!=ratioy){
      if(ratiox<ratioy){
        ratioadd=(ratioy-ratiox)/4;
        ratioy=ratiox+ratioadd;
        }else{
        ratioadd=(ratiox-ratioy)/4;
        ratiox=ratioy+ratioadd;
      }
    }
    ofsx=(256-(w*ratiox))/2;
    ofsy=(256-(h*ratioy))/2;
    _consolePrintf("ratio=%f,%f, radd=%f, ofs=%f, %f.\n",ratiox,ratioy,ratioadd,ofsx,ofsy);
  }

  pDstModel->wc=pSrcModel->wc;
  pDstModel->LinesCount=pSrcModel->LinesCount;
  pDstModel->pLines=(TModelLine*)safemalloc(&MM_SKKOSKHW,sizeof(TModelLine)*pDstModel->LinesCount);
  for(u32 lidx=0;lidx<pDstModel->LinesCount;lidx++){
    TModelLine *psrcline=&pSrcModel->pLines[lidx];
    TModelLine *pdstline=&pDstModel->pLines[lidx];
    pdstline->TotalLen=psrcline->TotalLen;
    pdstline->fx=(psrcline->fx-minx)*ratiox+ofsx;
    pdstline->fy=(psrcline->fy-miny)*ratioy+ofsy;
    pdstline->PointsCount=psrcline->PointsCount;
    pdstline->pPoints=(TModelLinePoint*)safemalloc(&MM_SKKOSKHW,sizeof(TModelLinePoint)*pdstline->PointsCount);
    for(u32 pidx=0;pidx<pdstline->PointsCount;pidx++){
      TModelLinePoint *psrcpoi=&psrcline->pPoints[pidx];
      TModelLinePoint *pdstpoi=&pdstline->pPoints[pidx];
      pdstpoi->x=psrcpoi->x*ratiox;
      pdstpoi->y=psrcpoi->y*ratioy;
      pdstpoi->NormalLen=psrcpoi->NormalLen;
    }
  }

  for(u32 lidx=0;lidx<pDstModel->LinesCount;lidx++){
    ModelLine_CreateLength(&pDstModel->pLines[lidx]);
  }
  
  return(KogakiFlag);
}

static void Model_Draw(const TModel *pModel,CglCanvas *pcan,double ofsx,double ofsy,double ratio,u16 col)
{
  for(u32 lidx=0;lidx<pModel->LinesCount;lidx++){
    TModelLine *pline=&pModel->pLines[lidx];
    double x=ofsx+(pline->fx*ratio);
    double y=ofsy+(pline->fy*ratio);
    double lx=x,ly=y;
    for(u32 pidx=0;pidx<pline->PointsCount;pidx++){
      TModelLinePoint *ppoi=&pline->pPoints[pidx];
      x+=ppoi->x*ratio;
      y+=ppoi->y*ratio;
      DrawLine(pcan,lx,ly,x,y,col);
      lx=x; ly=y;
    }
  }
}

static void Models_AddRefer(const TModel *pSrcModel,UnicodeChar wc)
{
  _consolePrintf("AddRefer $%04x to %d.\n",wc,ModelsCount);
  
  pModels=(TModel*)saferealloc(&MM_SKKOSKHW_Model,pModels,sizeof(TModel)*(ModelsCount+1));
  TModel *pmdl=&pModels[ModelsCount++];
  
  pmdl->wc=wc;
  pmdl->LinesCount=pSrcModel->LinesCount;
  pmdl->pLines=(TModelLine*)safemalloc(&MM_SKKOSKHW_Model,sizeof(TModelLine)*pmdl->LinesCount);
//  _consolePrintf("LinesCount: %d.\n",pmdl->LinesCount);
  for(u32 lidx=0;lidx<pmdl->LinesCount;lidx++){
    TModelLine *pline=&pmdl->pLines[lidx];
    pline->fx=pSrcModel->pLines[lidx].fx;
    pline->fy=pSrcModel->pLines[lidx].fy;
    pline->PointsCount=pSrcModel->pLines[lidx].PointsCount;
//    _consolePrintf("First= %f, %f, PointsCount:%d.\n",pline->fx,pline->fy,pline->PointsCount);
    pline->pPoints=(TModelLinePoint*)safemalloc(&MM_SKKOSKHW_Model,sizeof(TModelLinePoint)*pline->PointsCount);
    for(u32 pidx=0;pidx<pline->PointsCount;pidx++){
      TModelLinePoint *ppoi=&pline->pPoints[pidx];
      ppoi->x=pSrcModel->pLines[lidx].pPoints[pidx].x;
      ppoi->y=pSrcModel->pLines[lidx].pPoints[pidx].y;
//      _consolePrintf("Points %d: %f, %f.\n",pidx,ppoi->x,ppoi->y);
    }
    ModelLine_CreateLength(pline);
  }
  
  Models_SaveToBin();
}

// ---------------------------------------------------------------------------

static void Models_Free(void)
{
  if(pModels!=NULL){
    for(u32 midx=0;midx<ModelsCount;midx++){
      TModel *pmdl=&pModels[midx];
      for(u32 lidx=0;lidx<pmdl->LinesCount;lidx++){
        TModelLine *pline=&pmdl->pLines[lidx];
        pline->PointsCount=0;
        if(pline->pPoints!=NULL){
          safefree(&MM_SKKOSKHW_Model,pline->pPoints); pline->pPoints=NULL;
        }
      }
      pmdl->LinesCount=0;
      if(pmdl->pLines!=NULL){
        safefree(&MM_SKKOSKHW_Model,pmdl->pLines); pmdl->pLines=NULL;
      }
    }
    ModelsCount=0;
    safefree(&MM_SKKOSKHW_Model,pModels); pModels=NULL;
  }
}

// ---------------------------------------------------------------------------

